"""Module for login widgets."""

import typing

from PySide6.QtCore import QDate, Qt
from PySide6.QtGui import QKeyEvent
from PySide6.QtWidgets import (
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QSizePolicy,
    QWidget,
)

from fwtv.widgets import settings_widget


class LoginWidget(QWidget):
    """Provide a way to login for the user."""

    def __init__(self, *args: typing.Any, **kwargs: typing.Any):
        super().__init__(*args, **kwargs)
        self.qh = QHBoxLayout(self)

        self.fetch_data_label = QLabel('Fetch data', self)
        self.fetch_data_label.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        self.qh.addWidget(self.fetch_data_label)

        # defaults to fetching data from the last 2 months in order to not fetch entire data history
        self.start_picker = settings_widget.DateSettingWidget('From', QDate.currentDate().addMonths(-2), self)
        self.qh.addWidget(self.start_picker)

        self.end_picker = settings_widget.DateSettingWidget('Until', QDate.currentDate(), self)
        self.qh.addWidget(self.end_picker)

        self.label = QLabel('Enter api key', self)
        self.label.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        self.qh.addWidget(self.label)

        self.key = QLineEdit(self)
        self.key.setEchoMode(QLineEdit.EchoMode.Password)
        self.qh.addWidget(self.key)

        self.button = QPushButton('Fetch data', self)
        self.button.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        self.qh.addWidget(self.button)

        self.timeout = settings_widget.IntegerPickerWidget('Timeout', 5)
        self.qh.addWidget(self.timeout)

        self.setLayout(self.qh)

    def keyPressEvent(self, event: QKeyEvent):  # noqa: N802
        """Handle key press events."""
        super().keyPressEvent(event)
        if event.key() == Qt.Key.Key_Return:
            self.button.clicked.emit()



"""Module for setting widgets."""

import typing

from PySide6.QtCore import QDate
from PySide6.QtWidgets import (
    QComboBox,
    QDateEdit,
    QDoubleSpinBox,
    QHBoxLayout,
    QLabel,
    QSizePolicy,
    QWidget,
)


class TeamOrEmployeeSettingWidget(QWidget):
    """A widget to select a team or employee."""

    def __init__(self, *args: typing.Any, **kwargs: typing.Any):
        super().__init__(*args, **kwargs)
        self.qh = QHBoxLayout(self)
        self.label = QLabel('Select a team or an employee', self)
        self.qh.addWidget(self.label)

        self.selector = QComboBox(self)
        self.selector.setEditable(True)
        self.selector.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        self.qh.addWidget(self.selector)

        self.setLayout(self.qh)


class DateSettingWidget(QWidget):
    """A picker for dates."""

    def __init__(self, label: str, date: QDate, *args: typing.Any, **kwargs: typing.Any):
        super().__init__(*args, **kwargs)
        self.qh = QHBoxLayout(self)
        self.label = QLabel(label, self)
        self.qh.addWidget(self.label)

        self.date = QDateEdit(date, self)
        self.date.setCalendarPopup(True)
        self.date.setDisplayFormat('yyyy-MM-dd')
        self.qh.addWidget(self.date)

        self.setLayout(self.qh)


class IntegerPickerWidget(QWidget):
    """A picker for integers."""

    def __init__(self, label: str, default: int, *args: typing.Any, **kwargs: typing.Any):
        super().__init__(*args, **kwargs)
        self.qh = QHBoxLayout(self)
        self.label = QLabel(label, self)
        self.qh.addWidget(self.label)

        self.picker = QDoubleSpinBox(self, maximum=999)  # allow maximum of 9-digit number
        self.picker.setSingleStep(1)
        self.picker.setDecimals(0)
        self.picker.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)
        self.picker.setValue(default)
        self.qh.addWidget(self.picker)

        self.setLayout(self.qh)

    def value(self) -> int:
        """Get the value of the integer picker."""
        return int(self.picker.value())


class SettingsWidget(QWidget):
    """Provide some settings for the user."""

    def __init__(self, *args: typing.Any, **kwargs: typing.Any):
        super().__init__(*args, **kwargs)
        self.qh = QHBoxLayout(self)
        self.team_selector = TeamOrEmployeeSettingWidget(self)
        self.qh.addWidget(self.team_selector)

        last_month = QDate.currentDate().addMonths(-1)
        self.start_picker = DateSettingWidget('Start on', QDate(last_month.year(), last_month.month(), 1), self)
        self.qh.addWidget(self.start_picker)

        self.end_picker = DateSettingWidget(
            'End on',
            QDate(last_month.year(), last_month.month(), last_month.daysInMonth()).addDays(1),
            self,
        )
        self.qh.addWidget(self.end_picker)

        self.tolerance_selector = IntegerPickerWidget('Tolerance', 1)
        self.qh.addWidget(self.tolerance_selector)

        self.setLayout(self.qh)



"""Entrypoint and main window of the application."""

import datetime
import sys
import typing

import factorialhr
from PySide6.QtWidgets import QApplication, QMessageBox, QVBoxLayout, QWidget

import fwtv
from fwtv.objects import async_converter
from fwtv.widgets import login_widget, working_time_widget


class MainWindow(QWidget):
    """Main window widget."""

    def __init__(self, *args: typing.Any, **kwargs: typing.Any):
        super().__init__(*args, **kwargs)
        self.setWindowTitle(f'Factorial working time verification - version {fwtv.__version__}')
        self.qv = QVBoxLayout()
        self.login = login_widget.LoginWidget(self)
        self.qv.addWidget(self.login)

        self.verification_widget = working_time_widget.WorkingTimeWidget(self)
        self.qv.addWidget(self.verification_widget)

        self.setLayout(self.qv)

        self.login.button.clicked.connect(async_converter.ToAsync(self.fetch_data))

    async def fetch_data(self):
        """Fetch data from the api."""
        self.login.button.hide()
        async with factorialhr.ApiClient(auth=factorialhr.ApiKeyAuth(self.login.key.text())) as api:
            try:
                _attendances = await factorialhr.ShiftEndpoint(api).all(
                    start_on=typing.cast(
                        datetime.date | None,
                        self.login.start_picker.date.date().toPython(),
                    ),
                    end_on=typing.cast(
                        datetime.date | None,
                        self.login.end_picker.date.date().toPython(),
                    ),
                    timeout=self.login.timeout.value(),
                )
                _employees = await factorialhr.EmployeeEndpoint(api).all()
                _teams = await factorialhr.TeamEndpoint(api).all()
            except Exception as e:  # noqa: BLE001
                message_box = QMessageBox(self)
                message_box.setIcon(QMessageBox.Icon.Critical)
                message_box.setText(f'{type(e).__name__}\n{e}')
                message_box.setStandardButtons(QMessageBox.StandardButton.Ok)
                message_box.setDefaultButton(QMessageBox.StandardButton.Ok)
                message_box.exec()
                return
            finally:
                self.login.button.show()
        self.verification_widget.set_data(_teams, _attendances, _employees)


def main() -> int:
    """Entrypoint for application."""
    app = QApplication(sys.argv)
    window = MainWindow()
    window.resize(1200, 675)
    window.show()
    return app.exec()


if __name__ == '__main__':
    sys.exit(main())



"""Test for the attendance verification module."""

import datetime
import uuid

import pytest

from fwtv import verifier


@pytest.mark.parametrize(
    ('clock_in', 'clock_out'),
    [
        (
            datetime.datetime(2023, 1, 1, 0, 0, 0, 0, tzinfo=datetime.UTC),
            datetime.datetime(2023, 1, 1, 0, 0, 0, 0, tzinfo=datetime.UTC),
        ),
        (
            datetime.datetime(2023, 1, 1, 0, 0, 0, 1, tzinfo=datetime.UTC),
            datetime.datetime(2023, 1, 1, 0, 0, 0, 0, tzinfo=datetime.UTC),
        ),
    ],
)
def test_clock_in_earlier_than_clock_out(clock_in: datetime.datetime, clock_out: datetime.datetime):
    """Test whether an error is raised if clock_out is before clock_in."""
    with pytest.raises(ValueError):  # noqa: PT011
        verifier.Attendance(clock_in, clock_out)


@pytest.mark.parametrize(
    ('clock_in', 'clock_out'),
    [
        (
            datetime.datetime(2023, 1, 1, 0, 0, 0, 0, tzinfo=datetime.UTC),
            datetime.datetime(2023, 1, 1, 0, 0, 0, 1, tzinfo=datetime.UTC),
        )
    ],
)
def test_str_and_repr(clock_in: datetime.datetime, clock_out: datetime.datetime):
    """Test whether the str and repr at as expected."""
    expected = f'Attendance(clock_in={clock_in!s}, clock_out={clock_out!s})'
    str_ = str(verifier.Attendance(clock_in, clock_out))
    assert str_ == expected
    repr_ = repr(verifier.Attendance(clock_in, clock_out))
    assert repr_ == expected


single_attendance = [
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 1, 8, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 1, 10, tzinfo=datetime.UTC),
    )
]
normal_attendances = [
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 1, 8, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 1, 10, tzinfo=datetime.UTC),
    ),
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 1, 11, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 1, 14, tzinfo=datetime.UTC),
    ),
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 1, 15, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 1, 20, tzinfo=datetime.UTC),
    ),
]
attendances_with_intersection = [
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 1, 8, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 1, 10, tzinfo=datetime.UTC),
    ),
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 1, 9, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 1, 14, tzinfo=datetime.UTC),
    ),
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 1, 13, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 1, 20, tzinfo=datetime.UTC),
    ),
]
begin_interception = [
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 1, 8, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 1, 10, tzinfo=datetime.UTC),
    ),
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 1, 9, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 1, 10, tzinfo=datetime.UTC),
    ),
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 1, 13, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 1, 17, tzinfo=datetime.UTC),
    ),
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 1, 13, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 1, 20, tzinfo=datetime.UTC),
    ),
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 1, 14, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 1, 15, tzinfo=datetime.UTC),
    ),
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 1, 15, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 1, 16, tzinfo=datetime.UTC),
    ),
]
attendances_with_inception = [
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 1, 8, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 1, 10, tzinfo=datetime.UTC),
    ),
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 1, 9, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 1, 10, tzinfo=datetime.UTC),
    ),
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 1, 13, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 1, 17, tzinfo=datetime.UTC),
    ),
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 1, 13, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 1, 20, tzinfo=datetime.UTC),
    ),
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 1, 14, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 1, 15, tzinfo=datetime.UTC),
    ),
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 1, 15, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 1, 16, tzinfo=datetime.UTC),
    ),
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 1, 15, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 1, 21, tzinfo=datetime.UTC),
    ),
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 1, 20, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 1, 22, tzinfo=datetime.UTC),
    ),
]
multiple_days = [
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 1, 8, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 2, 2, tzinfo=datetime.UTC),
    ),
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 2, 18, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 3, 19, tzinfo=datetime.UTC),
    ),
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 3, 13, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 4, 5, tzinfo=datetime.UTC),
    ),
]
random_order = [
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 15, 8, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 15, 10, tzinfo=datetime.UTC),
    ),
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 14, 9, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 14, 10, tzinfo=datetime.UTC),
    ),
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 20, 13, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 20, 17, tzinfo=datetime.UTC),
    ),
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 1, 13, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 1, 20, tzinfo=datetime.UTC),
    ),
    verifier.Attendance(
        clock_in=datetime.datetime(2022, 12, 31, 14, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2022, 12, 31, 15, tzinfo=datetime.UTC),
    ),
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 1, 15, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 1, 16, tzinfo=datetime.UTC),
    ),
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 2, 4, 15, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 2, 4, 21, tzinfo=datetime.UTC),
    ),
    verifier.Attendance(
        clock_in=datetime.datetime(2023, 1, 1, 20, tzinfo=datetime.UTC),
        clock_out=datetime.datetime(2023, 1, 1, 22, tzinfo=datetime.UTC),
    ),
]


@pytest.mark.parametrize(
    ('attendances', 'expected'),
    [
        ([], datetime.timedelta()),
        (single_attendance, datetime.timedelta(hours=2)),
        (normal_attendances, datetime.timedelta(hours=10)),
        (attendances_with_intersection, datetime.timedelta(hours=12)),
        (attendances_with_inception, datetime.timedelta(hours=11)),
        (multiple_days, datetime.timedelta(hours=53)),
        (random_order, datetime.timedelta(hours=23)),
    ],
)
def test_calculate_attended_time(attendances: list[verifier.Attendance], expected: datetime.timedelta):
    """Test the calculated attended time."""
    actual = verifier.calculate_time_attended(attendances)
    assert actual == expected


@pytest.mark.parametrize(
    ('attendances', 'expected'),
    [
        ([], datetime.timedelta()),
        (single_attendance, datetime.timedelta(hours=0)),
        (normal_attendances, datetime.timedelta(hours=2)),
        (begin_interception, datetime.timedelta(hours=3)),
        (attendances_with_intersection, datetime.timedelta(hours=0)),
        (attendances_with_inception, datetime.timedelta(hours=3)),
        (multiple_days, datetime.timedelta(hours=16)),
        (random_order, datetime.timedelta(hours=824)),
    ],
)
def test_calculate_break_time(attendances: list[verifier.Attendance], expected: datetime.timedelta):
    """Test the calculated break time."""
    actual = verifier.calculate_break_time(attendances)
    assert actual == expected


@pytest.mark.parametrize(
    ('attendances', 'expected'),
    [
        (single_attendance, {datetime.date(2023, 1, 1)}),
        (normal_attendances, {datetime.date(2023, 1, 1)}),
        (attendances_with_intersection, {datetime.date(2023, 1, 1)}),
        (attendances_with_inception, {datetime.date(2023, 1, 1)}),
        (
            multiple_days,
            {
                datetime.date(2023, 1, 1),
                datetime.date(2023, 1, 2),
                datetime.date(2023, 1, 3),
                datetime.date(2023, 1, 4),
            },
        ),
        (
            random_order,
            {
                datetime.date(2023, 1, 1),
                datetime.date(2022, 12, 31),
                datetime.date(2023, 1, 15),
                datetime.date(2023, 1, 14),
                datetime.date(2023, 1, 20),
                datetime.date(2023, 2, 4),
            },
        ),
    ],
)
def test_error_affected_days(attendances: list[verifier.Attendance], expected: datetime.date):
    """Test whether the found error is as expected for affected days."""
    error = verifier.Error(reason=uuid.uuid4().hex, attendances=attendances)
    assert error.days_affected == expected


@pytest.mark.parametrize(
    ('attendances', 'expected'),
    [
        ([], datetime.timedelta()),
        (single_attendance, datetime.timedelta(hours=2)),
        (normal_attendances, datetime.timedelta(hours=10)),
        (attendances_with_intersection, datetime.timedelta(hours=12)),
        (attendances_with_inception, datetime.timedelta(hours=11)),
        (multiple_days, datetime.timedelta(hours=53)),
        (random_order, datetime.timedelta(hours=23)),
    ],
)
def test_error_attended_time(attendances: list[verifier.Attendance], expected: datetime.timedelta):
    """Test whether the found error is as expected for attended time."""
    error = verifier.Error(reason=uuid.uuid4().hex, attendances=attendances)
    assert error.time_attended == expected


@pytest.mark.parametrize(
    ('attendances', 'expected'),
    [
        ([], datetime.timedelta()),
        (single_attendance, datetime.timedelta(hours=0)),
        (normal_attendances, datetime.timedelta(hours=2)),
        (begin_interception, datetime.timedelta(hours=3)),
        (attendances_with_intersection, datetime.timedelta(hours=0)),
        (attendances_with_inception, datetime.timedelta(hours=3)),
        (multiple_days, datetime.timedelta(hours=16)),
        (random_order, datetime.timedelta(hours=824)),
    ],
)
def test_error_break_time(attendances: list[verifier.Attendance], expected: datetime.timedelta):
    """Test whether the found error is as expected for break time."""
    error = verifier.Error(reason=uuid.uuid4().hex, attendances=attendances)
    assert error.break_time == expected


@pytest.mark.parametrize(
    ('attendances', 'error_count'),
    [
        (single_attendance, 0),
        (normal_attendances, 0),
        (attendances_with_intersection, 1),
        (attendances_with_inception, 1),
        (multiple_days, 3),
        (random_order, 2),
    ],
)
def test_verify_work_time(attendances: list[verifier.Attendance], error_count: int):
    """Verify found error count."""
    errors = verifier.verify_attendances(attendances, tolerance=datetime.timedelta())
    assert len(errors) == error_count



"""A wrapper to call async functions.

Taken from https://doc.qt.io/qtforpython-6.5/examples/example_async_minimal.html
"""

import traceback
import typing

import outcome
import trio
from PySide6.QtCore import QEvent, QObject
from PySide6.QtWidgets import QApplication


class ToAsync(QObject):  # noqa: D101
    class ReenterQtObject(QObject):
        """This is a QObject to which an event will be posted, allowing Trio to resume when the event is handled.

        event.fn() is the next entry point of the Trio event loop.
        """  # noqa: D404

        def event(self, event) -> bool:  # noqa: ANN001, D102
            if event.type() == QEvent.User + 1:  # pyright: ignore [reportAttributeAccessIssue]
                event.fn()
                return True
            return False

    class ReenterQtEvent(QEvent):
        """This is the QEvent that will be handled by the ReenterQtObject.

        self.fn is the next entry point of the Trio event loop.
        """  # noqa: D404

        def __init__(self, fn):  # noqa: ANN001
            super().__init__(QEvent.Type(QEvent.User + 1))  # pyright: ignore [reportAttributeAccessIssue]
            self.fn = fn

    def __init__(self, signal: typing.Callable):
        super().__init__()
        self.reenter_qt = self.ReenterQtObject()
        self.entry = signal

    def __call__(self, *args, **kwargs):  # noqa: ANN003, ANN002, D102, ARG002
        trio.lowlevel.start_guest_run(
            self.entry,
            *args,
            run_sync_soon_threadsafe=self.next_guest_run_schedule,
            done_callback=self.trio_done_callback,
        )

    def next_guest_run_schedule(self, fn):  # noqa: ANN001
        """This function serves to re-schedule the guest (Trio) event loop inside the host (Qt) event loop.

        It is called by Trio at the end of an event loop run in order to relinquish back to Qt's event loop.
        By posting an event on the Qt event loop that contains Trio's next entry point, it ensures that Trio's
        event loop will be scheduled again by Qt.
        """  # noqa: D401, D404
        QApplication.postEvent(self.reenter_qt, self.ReenterQtEvent(fn))

    def trio_done_callback(self, outcome_):  # noqa: ANN001
        """This function is called by Trio when its event loop has finished."""  # noqa: D401, D404
        if isinstance(outcome_, outcome.Error):
            error = outcome_.error
            traceback.print_exception(type(error), error, error.__traceback__)



"""Module for table widgets."""

import collections
import typing

import tabulate
from PySide6.QtCore import Qt
from PySide6.QtGui import QClipboard, QKeyEvent
from PySide6.QtWidgets import (
    QHeaderView,
    QLabel,
    QSizePolicy,
    QTableView,
    QTableWidget,
    QTableWidgetItem,
    QVBoxLayout,
    QWidget,
)

from fwtv import verifier


class ErrorTable(QTableWidget):
    """Display errors to user."""

    def __init__(self, labels: list[str], *args: typing.Any, **kwargs: typing.Any):
        self.labels = labels
        super().__init__(0, len(self.labels), *args, **kwargs)
        self.setHorizontalHeaderLabels(self.labels)
        self.setSelectionBehavior(QTableView.SelectionBehavior.SelectRows)
        self._set_data({})

    def _set_data(self, data: dict[str, list[list[str]]]):
        self.setSortingEnabled(False)
        self.setRowCount(sum(len(failures) for failures in data.values()))
        row_index = 0
        for name, errors in data.items():
            for error in errors:
                self._set_row(row_index, name, *error)
                row_index += 1
        self.horizontalHeader().resizeSections(QHeaderView.ResizeMode.ResizeToContents)
        self.setSortingEnabled(True)

    def _set_row(self, row_index: int, *columns: str):
        for column_index, text in enumerate(columns):
            column_item = QTableWidgetItem(text)
            column_item.setFlags(
                Qt.ItemFlag.ItemNeverHasChildren
                | Qt.ItemFlag.ItemIsAutoTristate
                | Qt.ItemFlag.ItemIsEnabled
                | Qt.ItemFlag.ItemIsSelectable
            )
            self.setItem(row_index, column_index, column_item)

    def keyPressEvent(self, event: QKeyEvent):  # noqa: N802
        """Handle key press events."""
        super().keyPressEvent(event)
        if event.key() == Qt.Key.Key_C and (event.modifiers() & Qt.KeyboardModifier.ControlModifier):
            row_by_index = collections.defaultdict(list)
            for index in self.selectedIndexes():
                row_by_index[index.row()].append(index.data())
            table = tabulate.tabulate(row_by_index.values(), headers=self.labels, tablefmt='orgtbl')
            QClipboard().setText(table)


class PreconditionErrorsTableWidget(ErrorTable):
    """Display precondition errors to user."""

    def __init__(self, *args: typing.Any, **kwargs: typing.Any):
        super().__init__(['Name', 'Error'], *args, **kwargs)
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

    def set_data(self, data: dict[str, list[list[str]]]):
        """Set precondition data."""
        super()._set_data(data)
        self.horizontalHeader().setSectionResizeMode(1, QHeaderView.ResizeMode.Stretch)
        if not any(data) and not self.isHidden():
            self.hide()
        elif any(data) and self.isHidden():
            self.show()


class AdditionalInformationWidget(QWidget):
    """Display additional information to user."""

    def __init__(self, name: str, error: verifier.Error, *args: typing.Any, **kwargs: typing.Any):
        super().__init__(*args, **kwargs)
        self.qv = QVBoxLayout(self)

        self.label = QLabel(name)
        self.qv.addWidget(self.label)

        self.attendances = QTableWidget(len(error.attendances), 2, self)
        for i, a in enumerate(error.attendances):
            column_item = QTableWidgetItem(a.clock_in.strftime('%d.%m %H:%M'))
            column_item.setFlags(
                Qt.ItemFlag.ItemNeverHasChildren | Qt.ItemFlag.ItemIsAutoTristate | Qt.ItemFlag.ItemIsEnabled
            )
            self.attendances.setItem(i, 0, column_item)
            column_item = QTableWidgetItem(a.clock_out.strftime('%d.%m %H:%M'))
            column_item.setFlags(
                Qt.ItemFlag.ItemNeverHasChildren | Qt.ItemFlag.ItemIsAutoTristate | Qt.ItemFlag.ItemIsEnabled
            )
            self.attendances.setItem(i, 1, column_item)
        self.attendances.setHorizontalHeaderLabels(['Clock in', 'Clock out'])
        self.attendances.horizontalHeader().resizeSections(QHeaderView.ResizeMode.ResizeToContents)
        self.qv.addWidget(self.attendances)

        self.error = QLabel(error.reason)
        self.qv.addWidget(self.error)

        self.setLayout(self.qv)


class FailuresTableWidget(ErrorTable):
    """Display failures to user."""

    def __init__(self, *args: typing.Any, **kwargs: typing.Any):
        super().__init__(
            [
                'Name',
                'Affected Day(s)',
                'Reason',
                'Cumulated Break',
                'Cumulated Attendance',
            ],
            *args,
            **kwargs,
        )
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        self.data: dict[str, list[verifier.Error]] = {}
        self.cellDoubleClicked.connect(self.display_additional_information)

    def set_data(self, data: dict[str, list[verifier.Error]]):
        """Set verification errors in failures table."""
        self.data = data
        entries = collections.defaultdict(list)
        for name, error in data.items():
            for e in error:
                affected_days = [str(day) for day in e.days_affected]
                affected_days.sort()
                entries[name].append(
                    [
                        ', '.join(affected_days),
                        e.reason,
                        str(e.break_time),
                        str(e.time_attended),
                    ]
                )
        super()._set_data(entries)
        self.horizontalHeader().setSectionResizeMode(2, QHeaderView.ResizeMode.Stretch)

    def display_additional_information(self, row: int, _: int):
        """Display additional information to the user."""
        rows: list[tuple[str, verifier.Error]] = []
        for name, error in self.data.items():
            for e in error:
                rows.insert(0, (name, e))  # instead of reversing
        self.new_window = AdditionalInformationWidget(rows[row][0], rows[row][1])
        self.new_window.show()
